open Printf
open Language


(*
type wff =
  | False
  | P of string
  | Neg of wff
  | And of wff * wff
  | Or of wff * wff
  | Imp of wff * wff 
  | Iff of wff * wff 
*)	

(* representation of At-formulas *)

(*

type atm = Atm of int
(* internal representation of formulas based on atoms a(k) *)
   
type at_wff =
   | AtFalse
   | At of atm
   | AtP of string    
   | AtNeg of at_wff
   | AtAnd of at_wff * at_wff
   | AtOr of at_wff * at_wff
   | AtImp of at_wff * at_wff 
*)


(* string tranlsating an at_wff *)
let rec at_wff_to_string f = 
  match f with
  | AtFalse -> "bot"
  | At ( Atm k) -> "a(" ^ string_of_int k ^ ")"
  | AtP str -> str   
  | AtNeg f1 -> "neg(" ^ at_wff_to_string f1 ^ ")"
  | AtAnd(f1,f2) ->
     "and(" ^ at_wff_to_string f1 ^ "," ^  at_wff_to_string f2 ^ ")"
  | AtOr(f1,f2) ->
     "or(" ^ at_wff_to_string f1 ^ "," ^  at_wff_to_string f2 ^ ")"
  | AtImp(f1,f2) ->
     "imp(" ^ at_wff_to_string f1 ^ "," ^  at_wff_to_string f2 ^ ")"
  | AtIff(f1,f2) ->
     "iff(" ^ at_wff_to_string f1 ^ "," ^  at_wff_to_string f2 ^ ")"


(*
let rec wff_to_string f = 
  match f with
  | False -> "bot"
  | P str -> str   
  | Neg f1 -> "neg(" ^ wff_to_string f1 ^ ")"
  | And(f1,f2) ->
     "and(" ^ wff_to_string f1 ^ "," ^  wff_to_string f2 ^ ")"
  | Or(f1,f2) ->
     "or(" ^ wff_to_string f1 ^ "," ^  wff_to_string f2 ^ ")"
  | Imp(f1,f2) ->
     "imp(" ^ wff_to_string f1 ^ "," ^  wff_to_string f2 ^ ")"
*)
								   
type atm_def =
  {
    atom : atm;   (* atom  a(0), a(1), ...  *)
    atom_def: at_wff;  (* definition of the atom   *) 
    form : wff		(* formula corresponding to the atom *) 
  }  

		    
    
type goal_def =
  {
    goal_at    : at_wff ;
    def_list : atm_def list ;
    count   :  int;   (* number of atoms in the gal *)
    goal_wff : wff
  }

(* get the atm defining f in def_list, if this exists *)
let rec get_atm f def_list =
  match def_list with
  | [] -> None
  | def1 :: defs -> 	    
     if def1.form =f then Some def1
     else get_atm f defs

 (*   
     f: current formula
     k: index of the next new atom 
      defs. list of already defined 
Return a triple

  (df,new_k, new_def)  

where:
df: definition of f
new_k: update k
new_defs: update defs with new definitions 

*)
					  

let rec build_atom_list  f k defs  =
  match f with
  |  False ->  (AtFalse,k,defs)
  |  P str ->
      (
      match ( get_atm f defs ) with
      |  Some def  -> (At def.atom,k,defs)
      |  None ->
	  let new_atm =  Atm k in
	  let new_def = {atom = new_atm ; atom_def = AtP str ; form = f } in
	  ( At new_atm  , k + 1,  new_def ::  defs    ) 			  

      )		   
(*
  |  Neg ( Neg f1 ) ->
     (
       match ( get_atm f defs ) with
       |  Some def  ->    (At def.atom,k,defs)
       |  None ->
	  let (af1,k1,def1s) =  build_atom_list  f1 k defs in
	  let new_atm =  Atm k1 in
	  let new_def = {atom = new_atm ; atom_def = AtNeg (AtNeg af1) ; form = f } in
	  ( At new_atm  , k1 + 1,  new_def ::  def1s    ) 			  
     )	
 *)
  |  Neg f1 -> 
     (
       match ( get_atm f defs ) with
       |  Some def  ->    (At def.atom,k,defs)
       |  None ->
	  let (af1,k1,def1s) =  build_atom_list  f1 k defs in
	  let new_atm =  Atm k1 in
	  let new_def = {atom = new_atm ; atom_def = AtNeg af1 ; form = f } in
	  ( At new_atm  , k1 + 1,  new_def ::  def1s    ) 			  
     )	
  |  And(f1,f2) ->
      let (af1,k1,def1s) =  build_atom_list  f1 k defs in
      let (af2,k2,def2s) =  build_atom_list  f2 k1 def1s in
      (AtAnd(af1,af2), k2, def2s )
  |  Or(f1,f2) ->
      let (af1,k1,def1s) =  build_atom_list  f1 k defs in
      let (af2,k2,def2s) =  build_atom_list  f2 k1 def1s in
      (AtOr(af1,af2), k2, def2s )
   |  Imp(f1,f2) ->  					      
	(
	  match ( get_atm f defs ) with
	  |  Some def  ->    (At def.atom,k,defs)
	  |  None ->
	      let (af1,k1,def1s) =  build_atom_list  f1 k defs in
	      let (af2,k2,def2s) =  build_atom_list  f2 k1 def1s in
	      let new_atm =  Atm k2 in
	      let new_def = {atom = new_atm ; atom_def = AtImp(af1, af2) ; form = f } in
	      ( At new_atm   , k2 + 1,  new_def ::  def2s    ) 			  
	)						      
   |  Iff(f1,f2) ->  					      
	(
	  match ( get_atm f defs ) with
	  |  Some def  ->    (At def.atom,k,defs)
	  |  None ->
	      let (af1,k1,def1s) =  build_atom_list  f1 k defs in
	      let (af2,k2,def2s) =  build_atom_list  f2 k1 def1s in
	      let new_atm =  Atm k2 in
	      let new_def = {atom = new_atm ; atom_def = AtIff(af1, af2) ; form = f } in
	      ( At new_atm   , k2 + 1,  new_def ::  def2s    ) 			  
	)	


let rec replace_PV_by_atms f gdef =
  match f with
  | P str -> 
       let df = List.find ( fun d -> d.form = f  ) gdef.def_list in 
       At (df.atom)
  | False -> AtFalse	 
  | Neg f1 -> AtNeg (  replace_PV_by_atms f1 gdef  )
  | And(f1,f2) -> AtAnd (  replace_PV_by_atms f1 gdef, replace_PV_by_atms f2 gdef  )
  | Or(f1,f2) -> AtOr (  replace_PV_by_atms f1 gdef , replace_PV_by_atms f2 gdef  )
  | Imp(f1,f2) -> AtImp (  replace_PV_by_atms f1 gdef ,  replace_PV_by_atms f2 gdef  )
  | Iff(f1,f2) -> AtImp (  replace_PV_by_atms f1 gdef ,  replace_PV_by_atms f2 gdef  )

	  
let build_goal_def form =
  let (af, k, defs ) = build_atom_list  form 0 [] in
  { count  = k;   goal_at = af ;   def_list = defs ; goal_wff = form }

let print_count_atms k = printf  "count_atms(%d).\n"  k


let print_def { atom = Atm k  ; atom_def = atDef ; form = f } =
  match atDef with
  | AtP str ->  printf "at_to_PV( a(%d) , %s ).\n"  k str
  | _ ->  printf "def( a(%d) , %s ).\n"  k ( at_wff_to_string  atDef)
  
let print_defs defs = ignore ( List.map print_def  defs )
    
let print_goal gdef =
  printf "goal( %s ).\n"  ( at_wff_to_string  gdef.goal_at) 
  (*  printf "goal_wff( %s ).\n"   ( at_wff_to_string    ( replace_PV_by_atms gdef.goal_wff gdef))    *)
(* printf "goal_wff( %s ).\n"   ( wff_to_string    gdef.goal_wff ) *)
  
let print_goal_def gdef =
  print_count_atms gdef.count ;
  print_defs  gdef.def_list ;
  print_goal  gdef
  
	 
let build f =
  let gdef =  build_goal_def f in
  print_goal_def gdef
     
    
(*** TEST  ****)

let p0 =  build_goal_def ( P "p" )     
let p1 =  build_goal_def (And ( P "p" , P "q" ) )      
let p2 =  build_goal_def (And ( P "p" , P "p" ) )      
let p3 =  build_goal_def (Imp ( P "p" , P "q" ) )      
let p4 =  build_goal_def (Imp ( P "p" , P "p" ) )      


  
let h = Imp ( Imp(  Neg (Neg (P "p")), P "p")  , Or ( Neg ( Neg (P "p")), Neg (P "p") )) 
    
let scott =  Imp ( h,    Or ( Neg ( Neg (P "p")), Neg (P "p"))) 
let t1 = build_goal_def scott  



(* antiscott *)
let ast = Imp ( scott, Or (  Imp(  Neg (Neg (P "p")), P "p") , Neg (Neg (P "p"))))

let dum a b  =  Or ( Imp (a, b ) ,  Imp ( b, a ) )

let dum2 = dum (P "p") (P "q")
		  
let dum4 =Or ( dum ( P "p1")  ( P "p2" ) , dum ( P "q1")  ( P "q2" )    )

let d3 = Or( dum (P "p") (P "q") , Imp  ( Or( P "p", P "q"), And( P "p", P "q")   ) )

let syj208a = And( Or( P "p11", Neg ( Neg ( P "p12")) ),
		   And(  Or( P "p21", Neg ( Neg ( P "p22")) ),
			 Or( P "p31", Neg ( Neg ( P "p32")) ) ) )

let syj208c = Or( And( P "p11", P "p21"),
		  Or ( And( P "p11", P "p31"),
		       Or ( And( P "p21", P "p31"),
			  Or ( And( P "p12", P "p22"),
			     Or ( And( P "p12", P "p32"),
				      And( P "p22", P "p32")  )))))

let syj208 = Imp( syj208a , syj208c)
