open Printf
open Language

type operator = No_op | Neg_op|  And_op | Or_op | Imp_op | Iff_op  


       
let is_atomic f =
  match f with
  | False | P _ -> true
  | _ -> false	
	
let is_atomic_or_neg f =
  match f with
  | False | P _ | Neg _ -> true
  | _ -> false	
	
let main_operator f =
  match f with
  |  False |  P _ -> No_op 
  |  Neg f1 -> Neg_op 
  |  And(f1,f2) -> And_op 
  |  Or(f1,f2) -> Or_op
  |  Imp(f1,f2) -> Imp_op
  |  Iff(f1,f2) -> Iff_op
     
let main_operator_opt f =
  match f with
  |  False |  P _ -> None 
  |  Neg f1 -> Some Neg_op 
  |  And(f1,f2) -> Some And_op 
  |  Or(f1,f2) -> Some Or_op
  |  Imp(f1,f2) -> Some Imp_op
  |  Iff(f1,f2) -> Some Iff_op
	
	
let op_to_string op =
  match op with
  | Neg_op -> "~"  
  | And_op -> "&"
  | Or_op -> "v"
  | Imp_op -> "->"
  | Iff_op -> "<->"
  | _ -> ""	
	
	
let add_parenthesis s =   "(" ^ s ^ ")"
				      
let rec wff_to_string f =
  match f with
  |  False -> "#f"
  |  P s -> s
  |  Neg f1 ->
      let s1 = wff_to_string f1 in
      let op1 =  main_operator f1 in
      let ss1 = if List.mem op1 [ And_op ; Or_op ; Imp_op ; Iff_op ]  then  add_parenthesis s1 else s1 in
      ( op_to_string Neg_op ) ^ ss1
  |  And(f1,f2) -> 
      let s1 = wff_to_string f1 in
      let s2 = wff_to_string f2 in
      let op1 =  main_operator f1 in
      let ss1 = if List.mem op1 [ Or_op ; Imp_op ; Iff_op  ]  then  add_parenthesis s1 else s1 in
      let op2 = main_operator f2 in
      let ss2 =  if List.mem op2 [ Or_op ; Imp_op ; Iff_op ]  then  add_parenthesis s2 else s2 in
      ss1 ^ " " ^  ( op_to_string And_op ) ^ " "  ^ ss2
  |  Or(f1,f2) -> 
      let s1 = wff_to_string f1 in
      let s2 = wff_to_string f2 in
      let op1 =  main_operator f1 in
      let ss1 = if List.mem op1 [ And_op ; Imp_op ; Iff_op ]  then  add_parenthesis s1 else s1 in
      let op2 = main_operator f2 in
      let ss2 =  if List.mem op2 [ And_op ; Imp_op ; Iff_op ]  then  add_parenthesis s2 else s2 in
      ss1 ^ " " ^  ( op_to_string Or_op ) ^ " "  ^ ss2						     
  |  Imp(f1,f2) -> 
      let s1 = wff_to_string f1 in
      let s2 = wff_to_string f2 in
      let op1 =  main_operator f1 in
      let ss1 = if List.mem op1 [ Imp_op ; Iff_op ]  then  add_parenthesis s1 else s1 in
      let op2 = main_operator f2 in
      let ss2 =  if List.mem op2 [ Imp_op ; Iff_op ]  then  add_parenthesis s2 else s2 in
      ss1 ^ " " ^  ( op_to_string Imp_op ) ^ " "  ^ ss2		
  |  Iff(f1,f2) -> 
      let s1 = wff_to_string f1 in
      let s2 = wff_to_string f2 in
      let op1 =  main_operator f1 in
      let ss1 = if List.mem op1 [ Imp_op ; Iff_op ]  then  add_parenthesis s1 else s1 in
      let op2 = main_operator f2 in
      let ss2 =  if List.mem op2 [ Imp_op ; Iff_op ]  then  add_parenthesis s2 else s2 in
      ss1 ^ " " ^  ( op_to_string Iff_op ) ^ " "  ^ ss2							      
			
